import matplotlib
import customtkinter as ctk
import pandas as pd
import tkcalendar as tkc
from hydroecolstm.data.read_data import read_forecast_data
import tkinter as tk
import torch
from CTkListbox import CTkListbox
from CTkToolTip import CTkToolTip
from pandastable import Table
from hydroecolstm.data.read_config import read_config
from hydroecolstm.model.create_model import create_model
from hydroecolstm.interface.utility import (ToplevelWindow,
                                            plot_time_series,
                                            check_size,check_linestyle,
                                            check_alpha,check_line_plot,
                                            check_marker, check_color, 
                                            check_ylim, combine_forecast)

class ApplicationFrame(ctk.CTkScrollableFrame):
    def __init__(self, container=None, config=None, globalData=None):
        super().__init__(container)
        
        # setup the grid layout manager
        self.config = config
        self.globalData = globalData
        self.columnconfigure((0), weight=1)
        self.__create_widgets() 
        
    # create widgets for sidebar frame
    def __create_widgets(self): 
        # ---------------------------------------------
        # create tabs
        self.tabview = ctk.CTkTabview(master=self, width = 750, border_width=1.5, 
                                      fg_color = "transparent")
        self.tabview.pack(fill='both',expand=1)
        self.tabview.add("1. Inputs")
        self.tabview.tab("1. Inputs").grid_columnconfigure(0, weight=0) 
        self.tabview.tab("1. Inputs").grid_columnconfigure(1, weight=0)
        self.tabview.tab("1. Inputs").grid_columnconfigure(2, weight=0)
        self.tabview.tab("1. Inputs").grid_columnconfigure(3, weight=1)

        
        self.tabview.add("2. Plot outputs")
        self.tabview.tab("2. Plot outputs").grid_columnconfigure((0,1,2), weight=1)
        
        # ---------------------------------------------content of load data tab
        # -----------------------------------------------------------Column 1
        self.import_label = ctk.CTkLabel(self.tabview.tab("1. Inputs"), 
                                         text="1. Internal import")
        self.import_label.grid(row=0, column=0, padx = 10, pady=(5,5), sticky="w")
        self.import_checkvalue = ctk.IntVar(value=0)         
        self.import_checkbox = ctk.CTkCheckBox(self.tabview.tab("1. Inputs"), 
                                               text=" ",
                                               command=self.import_button_event, 
                                               variable=self.import_checkvalue)
        self.import_checkbox.grid(row=1, column=0, padx = 10, pady=(10,10), sticky="we")
        CTkToolTip(self.import_checkbox, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=500, 
                   message="Check this box to use the model and data created from steps 1 to 4")

        # -----------------------------------------------------------Column 2
        self.file_selection_label = ctk.CTkLabel(self.tabview.tab("1. Inputs"), 
                                         text="2. Input data files")
        self.file_selection_label.grid(row=0, column=1, padx = 10, pady=(5,5), sticky="w")

        self.load_train_test_config = ctk.CTkButton(self.tabview.tab("1. Inputs"), 
                                               anchor='w', 
                                               text="Select best config file", 
                                               command=self.get_config_file)
        
        CTkToolTip(self.load_train_test_config, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=500, 
                   message="Select the configuration .yml file")
        
        self.load_train_test_config.grid(row=1, column=1, padx = 10, pady=(2,2), sticky="w")  
        self.config_file_name = ctk.CTkLabel(self.tabview.tab("1. Inputs"), 
                                         text="No file was selected")
        self.config_file_name.grid(row=1, column=2, padx = 10, pady=(2,2), sticky="w")  
        
        self.dynamic_file_button = ctk.CTkButton(self.tabview.tab("1. Inputs"), 
                                               anchor='w', 
                                               text="Select dynamic data file", 
                                               command=self.get_dynamic_file_forecast)
        self.dynamic_file_button.grid(row=2, column=1, padx = 10, pady=(2,2), sticky="w")
        CTkToolTip(self.dynamic_file_button, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=500, 
                   message="Please see 1. Data processing for more detail")
        
        self.dynamic_file_name = ctk.CTkLabel(self.tabview.tab("1. Inputs"), 
                                         text="No file was selected")
        self.dynamic_file_name.grid(row=2, column=2, padx = 10, pady=(2,2), sticky="w") 
        
    
        self.static_file_button = ctk.CTkButton(self.tabview.tab("1. Inputs"), 
                                               anchor='w', 
                                               text="Select static data file", 
                                               command=self.get_static_file_forecast)
        self.static_file_button.grid(row=3, column=1, padx = 10, pady=(2,2), sticky="w")
        CTkToolTip(self.static_file_button, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=500, 
                   message="Please see 1. Data processing for more detail")
        
        self.static_file_name = ctk.CTkLabel(self.tabview.tab("1. Inputs"), 
                                         text="No file was selected")
        self.static_file_name.grid(row=3, column=2, padx = 10, pady=(2,2), sticky="w") 
        

        self.model_button = ctk.CTkButton(self.tabview.tab("1. Inputs"), 
                                               anchor='w', 
                                               text="Load trained model", 
                                               command=self.load_model_state_dicts)
        self.model_button.grid(row=4, column=1, padx = 10, pady=(2,2), sticky="w") 
        CTkToolTip(self.static_file_button, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=500, 
                   message="Select file contains the model state dicts .pt file")
        
        self.model_name = ctk.CTkLabel(self.tabview.tab("1. Inputs"), 
                                         text="No model was loaded")
        self.model_name.grid(row=4, column=2, padx = 10, pady=(2,2), sticky="w") 
        
        self.load_scaler_button = ctk.CTkButton(self.tabview.tab("1. Inputs"), 
                                               anchor='w', 
                                               text="Load x and y scalers", 
                                               command=self.load_scalers)
        self.load_scaler_button.grid(row=5, column=1, padx = 10, pady=(2,2), sticky="w") 
        self.load_scaler_label = ctk.CTkLabel(self.tabview.tab("1. Inputs"), 
                                         text="No scalers were loaded")
        self.load_scaler_label.grid(row=5, column=2, padx = 10, pady=(2,2), sticky="w") 
        CTkToolTip(self.static_file_button, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=500, 
                   message="Load the file contains scaler for input and target features. " +  
                   " When you save this project, this is also inside the data.pt file." +
                   " So you can load this data.pt file here")
        
        
        self.model_label = ctk.CTkLabel(self.tabview.tab("1. Inputs"), 
                                         text="3. Forecast period")
        self.model_label.grid(row=6, column=1, columnspan=2, padx = 10, pady=(40,5), sticky="w")
        self.start_forecast = tkc.DateEntry(self.tabview.tab("1. Inputs"), 
                                         date_pattern= 'yyyy-mm-dd', width = 25,
                                         year=1800, month=1, day=1, font=ctk.CTkFont(size=14))
        self.start_forecast.grid(row= 7, column=1, columnspan=2, padx=30, pady=10, sticky='w')
        self.end_forecast = tkc.DateEntry(self.tabview.tab("1. Inputs"), 
                                       date_pattern= 'yyyy-mm-dd', width = 25,
                                       year=2015, month=1, day=1, font=ctk.CTkFont(size=14))
        self.end_forecast.grid(row= 8, column=1, columnspan=2, padx=30, pady=10, sticky='w')   
        
        
        # ---------------------------------------------------------------Column 3
        self.object_id_label = ctk.CTkLabel(self.tabview.tab("1. Inputs"), 
                                 text="4. Select object_id for forecasting")
        self.object_id_label.grid(row=0, column=3, columnspan=1, 
                                  padx = 10, pady=(5,5), sticky="w")
        self.object_id_forecast = CTkListbox(master=self.tabview.tab("1. Inputs"), 
                                           multiple_selection=True, border_width=1.5,
                                           text_color="black")
        
        self.object_id_forecast.grid(row=1, column=3, rowspan=5, padx = 10, 
                                     pady=(10,10), sticky="we")

        self.run_label = ctk.CTkLabel(self.tabview.tab("1. Inputs"), 
                                        text="5. Run forecast")
        self.run_label.grid(row=6, column=3, padx = 10, pady=(5,5), sticky="w")
       
        self.run_button = ctk.CTkButton(self.tabview.tab("1. Inputs"), 
                                              anchor='w', 
                                              text="Run", 
                                              command=self.run_forecast)
        self.run_button.grid(row=7, column=3, padx = 10, pady=(5,5), sticky="w") 
        
        #---------------------------------------------------------2 Plot output
        self.object_id_label =\
            ctk.CTkLabel(self.tabview.tab("2. Plot outputs"),
                         text="1. Select object id and target feature for plot")
            
        self.object_id_label.grid(row=0, column=0, columnspan=3, sticky="w", padx=5)

        self.object_id=ctk.CTkTextbox(self.tabview.tab("2. Plot outputs"), 
                                        height=30, border_width=1.5)
        
        self.object_id.insert("0.0", "object_id") 
        self.object_id.grid(row=1, column=0, sticky="w", padx=5, pady=5)
        
        self.object_id.bind('<KeyRelease>', self.get_object_id)

        self.target_feature=ctk.CTkTextbox(self.tabview.tab("2. Plot outputs"), 
                                             height=30, border_width=1.5)
        
        self.target_feature.insert("0.0", "target_feature") 
        
        self.target_feature.grid(row=2, column=0, sticky="w", padx=5, pady=5)
        
        self.target_feature.bind('<KeyRelease>', self.get_target_feature)
        
        self.next_object_id_button=ctk.CTkButton(self.tabview.tab("2. Plot outputs"), 
                                                   anchor='w',
                                                   command=self.next_object_id, 
                                                   text="Next object id")
        
        self.next_object_id_button.grid(row=1, column=1, sticky="w", 
                                        padx=5, pady=5)
        
        self.next_target_feature_button =\
            ctk.CTkButton(self.tabview.tab("2. Plot outputs"), anchor='w',
                          command=self.next_target_feature,
                          text="Next target features")
            
        self.next_target_feature_button.grid(row=2, column=1, sticky="w", 
                                             padx=5, pady=5)
        
        # Plot button
        self.update_plot=ctk.CTkButton(self.tabview.tab("2. Plot outputs"), anchor='we', 
                                 command=self.plot_figure, 
                                 text="Plot (update plot)")
        
        self.update_plot.grid(row=1, column=2,  
                              sticky="we", padx=5, pady=5)
        
        # show all data
        self.show_all_data=ctk.CTkButton(self.tabview.tab("2. Plot outputs"), anchor='w', 
                                         command=self.show_all_data_event,
                                         text="Show data source")
        self.show_all_data.grid(row=2, column=2, sticky="we", padx=5, pady=5)
        CTkToolTip(self.show_all_data, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, justify="left", 
                   message='Click here to show all data')
        
        
        #-------------------------------------------------------2. Plot setting
        # check to load/unload
        self.plot_timeseries=ctk.IntVar(value=0)        
        self.plot_timeseries_checkbox=ctk.CTkCheckBox(self.tabview.tab("2. Plot outputs"), 
                                               text="Show plot settings",
                                               command= self.show_plot_timeseries_setting, 
                                               variable=self.plot_timeseries)
        
        self.plot_timeseries_checkbox.grid(row=3, column=0, padx=10, pady=(10, 0), sticky="w")
        
        self.plot_timeseries_frame=ctk.CTkFrame(self.tabview.tab("2. Plot outputs"),
                                              height=400, fg_color="transparent")
        self.plot_timeseries_frame.columnconfigure((0,1,2), weight=1)

        #----------------------------------------------------------------------
        self.common_settings=ctk.CTkLabel(self.plot_timeseries_frame,
                                         text="Common settings")   
        self.common_settings.grid(row=0, column=0, sticky="w", padx=5)
        
        self.plot_title=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="plot title")
        
        self.plot_title.grid(row=1, column=0, sticky="w", padx=5)
        
        
        self.x_label=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="xlabel")
        self.x_label.grid(row=2, column=0, sticky="w", padx=5)
        
        self.y_label=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="ylabel")
        self.y_label.grid(row=3, column=0, sticky="w", padx=5)
        
        self.ylim_up=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="ylim (upper)")
        CTkToolTip(self.ylim_up, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, justify="left",
                   message='Upper limit of the y axis (input numeric value)')
        
        self.ylim_up.grid(row=4, column=0, sticky="w", padx=5)
        
        self.observed_data_plot=ctk.CTkLabel(self.plot_timeseries_frame,
                                         text="observed data plot")
        self.observed_data_plot.grid(row=0, column=1, sticky="w", padx=5)
          
        self.lineplot=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="line (true or false)")
        self.lineplot.grid(row=1, column=1, sticky="w", padx=5, pady=5)
                
        self.color_obs=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="color")
        self.color_obs.grid(row=2, column=1, sticky="w", padx=5, pady=5)
        CTkToolTip(self.color_obs, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, 
                   justify="left", 
                   message=str(list(matplotlib.colors.cnames.keys())))
        
        self.alpha_obs=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="alpha")    
        self.alpha_obs.grid(row=3, column=1, sticky="w", padx=5, pady=5)
        CTkToolTip(self.alpha_obs, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, 
                   justify="left", message='tranparency level: input any ' +
                   'numeric values between 0 and 1')
        
        self.size_obs=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="size")        
        self.size_obs.grid(row=4, column=1, sticky="w", padx=5, pady=5)
        CTkToolTip(self.size_obs, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, justify="left", 
                   message='input positive numeric value for line thickness')
        
        self.linestyle_obs=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="linestyle")    
        CTkToolTip(self.linestyle_obs, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, 
                   justify="left", message='solid, dashed, dashdot, dotted')
        
        self.linestyle_obs.grid(row=5, column=1, sticky="w", padx=5, pady=5)
        self.marker_obs=ctk.CTkEntry(self.plot_timeseries_frame,
                                       placeholder_text="marker")
        self.marker_obs.grid(row=6, column=1, sticky="w", padx=5, pady=5)
        CTkToolTip(self.marker_obs, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, 
                   justify="left", message="., o, s, ^, v, +, x")
        
        self.label_obs=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="label")
        
        self.label_obs.grid(row=7, column=1, sticky="w", padx=5, pady=5)
        
        #----------------------------------------------------------------------
        self.simulated_data_plot=ctk.CTkLabel(self.plot_timeseries_frame,
                                         text="simulated data plot")
        self.simulated_data_plot.grid(row=0, column=3, sticky="w", padx=5)
        
        self.color_sim=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="color")
        self.color_sim.grid(row=2, column=3, sticky="w", padx=5)
        self.alpha_sim=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="alpha")
        self.alpha_sim.grid(row=3, column=3, sticky="w", padx=5)
        self.size_sim=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="size")
        self.size_sim.grid(row=4, column=3, sticky="w", padx=5)
        CTkToolTip(self.size_sim, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, justify="left", 
                   message='input positive numeric value for line thickness')
        
        self.linestyle_sim=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="linestype")
        self.linestyle_sim.grid(row=5, column=3, sticky="w", padx=5)
        CTkToolTip(self.linestyle_sim, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, 
                   justify="left", message='solid, dashed, dashdot, dotted')
        self.label_sim=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="label")
        self.label_sim.grid(row=7, column=3, sticky="w", padx=5)

    def show_all_data_event(self):
        try:
            dataframe = combine_forecast(self.globalData,
                                         self.globalData["y_column_name"])
            
            self.table = Table(tk.Toplevel(self), dataframe=dataframe, 
                               showtoolbar=True, showstatusbar=True)
            self.table.show()
        except:
            pass
        
    def get_config_file(self):
        
        try:
            # get file name
            file_name = ctk.filedialog.askopenfilename(title="Select configuration file (.yml format)", 
                                                       filetypes=(('yml files', '*.yml'),
                                                                  ('All files', '*.*')))
            
            # read configuration file
            config = read_config(file_name)
            
            # update label
            self.config_file_name.configure(text= '...' + file_name[-15:])
            
            # update configuration
            self.config = config
            
            tk.messagebox.showinfo(title="Message box", 
                                   message="All of your previous configurations " +
                                   "were replaced by the configurations in the " +
                                   "selected file")
            
        except:
            tk.messagebox.showinfo(title="Message box", 
                                   message="No thing was change due either " +
                                   "canceling file selection or unsuccessfully " +
                                   "read of the selected configuration file")
            
    def get_dynamic_file_forecast(self):
        
        try:
            # get file name
            file_name = ctk.filedialog.askopenfilename(title="Select dynamic data file (.csv format)", 
                                                       filetypes=(('yml files', '*.csv'),
                                                                  ('All files', '*.*')))
            
            # update label
            self.dynamic_file_name.configure(text= '...' + file_name[-15:])
            
            # update configuration
            self.config["dynamic_data_file_forecast"] = [file_name]
            
            # try to remove object_id
            try:
                self.object_id_forecast.delete(index=0, last='END')
            except:
                pass
            
            # read dynamic data file to get header namey (object_id)
            object_id = pd.read_csv(file_name, delimiter=",", header=0)
            object_id = list(pd.unique(object_id["object_id"]))
             
            # remove columns with object_id and time
            for i in object_id:
                self.object_id_forecast.insert('END', option=i)
            
        except:
            self.dynamic_file_name.configure(text= 'No file was selected')


    def get_static_file_forecast(self):
        
        try:
            # get file name
            file_name = ctk.filedialog.askopenfilename(title="Select static data file (.csv format)", 
                                                       filetypes=(('yml files', '*.csv'),
                                                                  ('All files', '*.*')))
            
            # update label
            self.static_file_name.configure(text= '...' + file_name[-15:])
            
            # update configuration
            self.config["static_data_file_forecast"] = [file_name]
            
            
        except:
            self.static_file_name.configure(text= 'No file was selected')
            
    def load_model_state_dicts(self):
        try:
                
            # Load model state dicts
            file_name = ctk.filedialog.askopenfilename(title="Select model state dicts (.pt format)", 
                                                       filetypes=(('yml files', '*.pt'),
                                                                  ('All files', '*.*')))
            
            self.globalData["model"] = create_model(self.config, 
                                                    torch.load(file_name))
            
            # Set model to eval mode (in this mode, dropout = 0, no normlization)
            self.globalData["model"].eval()
            
            # Update selected model name
            self.model_name.configure(text= '...' + file_name[-15:])
            
        except:
            tk.messagebox.showinfo(title="Error", 
                                   message="Cannot load model state dicts")
            
    # Load scalers
    def load_scalers(self):
        try:
            # Load model state dicts
            file_name = ctk.filedialog.askopenfilename(title="Select scaler file (.pt format)", 
                                                       filetypes=(('yml files', '*.pt'),
                                                                  ('All files', '*.*')))
            
            scalers = torch.load(file_name)
            self.globalData["x_scaler"] = scalers["x_scaler"]
            self.globalData["y_scaler"] = scalers["y_scaler"]
            
            # Update selected model name
            self.load_scaler_label.configure(text= '...' + file_name[-15:])
            
        except:
            tk.messagebox.showinfo(title="Error", 
                                   message="Cannot load scalers")        
        
    # Show plot time series setting
    def show_plot_timeseries_setting(self):
        if self.plot_timeseries.get() == 0:
            self.plot_timeseries_frame.grid_forget()
        else:
            self.plot_timeseries_frame.grid(row=4, column=0, columnspan=3,
                                      sticky="w", padx=(0,0), pady=(20,20))
            
    # Get dropout
    def import_button_event(self):
        
        if self.import_checkbox.get() == 1:
            
            # Deactive other buttons
            self.load_train_test_config.configure(state="disabled")
            self.dynamic_file_button.configure(state="disabled")
            self.static_file_button.configure(state="disabled")
            self.model_button.configure(state="disabled")
            self.load_scaler_button.configure(state="disabled")
            
            # First try to remove all items in object_id_forecast
            try:
                self.object_id_forecast.delete(index=0, last='END')
            except:
                pass
            
            # Add items to object_id_forecast from globalData
            try:
                object_id = self.globalData['object_id'].copy()
                for i in object_id:
                    self.object_id_forecast.insert('END', option=i)
            except:
                pass
            
            # Input dynamic and static data
            if "dynamic_data_file" in self.config.keys():
                self.config["dynamic_data_file_forecast"] = ["dynamic_data_file"]
            if "static_data_file" in self.config.keys():
                self.config["static_data_file_forecast"] = ["static_data_file"]
                
        else:
            # Activate other buttons
            self.load_train_test_config.configure(state="normal")
            self.dynamic_file_button.configure(state="normal")
            self.static_file_button.configure(state="normal")
            self.model_button.configure(state="normal")
            self.load_scaler_button.configure(state="normal")
            
                
        
    def run_forecast(self):
        
        try:
            
            
            # Get forecast period
            self.config['forecast_period'] = pd.to_datetime(
                [self.start_forecast.get_date(), self.end_forecast.get_date()],
                format = "%Y-%m-%d %H:%M")

            # Get forecast id
            all_items = self.object_id_forecast.get(index='all')
            select_index = self.object_id_forecast.curselection()
            object_id_forecast = [all_items[i] for i in select_index]
            self.config['object_id_forecast'] = object_id_forecast

            # Add forecast data to globalData
            self.globalData.update(read_forecast_data(self.config))
 
            # Scale forecast data
            self.globalData["x_forecast_scale"] =\
                self.globalData["x_scaler"].transform(x=self.globalData["x_forecast"])

            # Run forward model
            y_forecast_scale_simulated =\
                self.globalData["model"].evaluate(self.globalData["x_forecast_scale"])

            tk.messagebox.showinfo(title="Message box", 
                                   message="Finished forward run")
            
            # Inverse transform y forcast to original scale
            self.globalData["y_forecast_simulated"] =\
                self.globalData["y_scaler"].inverse(y_forecast_scale_simulated)
            
        except:
            tk.messagebox.showinfo(title="Message box", 
                                   message="Error: Cannot run with forecast data")

    # Get dropout
    def next_object_id(self):
        try:
            if self.globalData["object_id_forecast_no"] > len(self.config["object_id_forecast"]) - 1:
                self.globalData["object_id_forecast_no"] = 0
            
            #print(self.globalData["object_id_forecast_no"])
            self.object_id.delete("0.0", "end")
            self.object_id.insert("0.0", self.config["object_id_forecast"]
                                  [self.globalData["object_id_forecast_no"]])
            self.globalData["object_id_forecast_plot"] = str(self.config["object_id_forecast"]
                                                    [self.globalData["object_id_forecast_no"]])
                
            self.globalData["object_id_forecast_no"] += 1
        except:
            None

    def next_target_feature(self):
        try:
            if self.globalData["target_feature_forecast_no"] >\
                len(self.config["target_features"]) - 1:
                self.globalData["target_feature_forecast_no"] = 0
               
            self.target_feature.delete("0.0", "end")
            
            self.target_feature.insert("0.0", self.config["target_features"]
                                       [self.globalData["target_feature_forecast_no"]])
            
            self.globalData["target_feature_forecast_plot"] =\
                str(self.config["target_features"]\
                    [self.globalData["target_feature_forecast_no"]])
                
            self.globalData["target_feature_forecast_no"] += 1
        except:
            None
            
    def get_object_id(self, dummy):
        self.globalData["object_id_forecast_plot"] =\
            str(self.object_id.get("0.0", "end"))
            
        self.globalData["object_id_forecast_plot"] =\
            self.globalData["object_id_forecast_plot"].strip()
            
        print(f"Selected object_id for plot =\
              {self.globalData['object_id_forecast_plot']}")

    def get_target_feature(self, dummy):
        self.globalData["target_feature_forecast_plot"] =\
            str(self.target_feature.get("0.0", "end"))
        self.globalData["target_feature_forecast_plot"] =\
            self.globalData["target_feature_forecast_plot"].strip()
        print(f"Selected target_feature for plot =\
              {self.globalData['target_feature_forecast_plot']}")
              
    def plot_figure(self):
        try:
            # Get information for plot
            key=self.globalData["object_id_forecast_plot"]
            idx=self.config["target_features"].index(
                self.globalData["target_feature_forecast_plot"])
            
            # General setting for plotting observed data
            lineplot=check_line_plot(self.lineplot)
            color_obs=check_color(self.color_obs, "coral")
            alpha_obs=check_alpha(self.alpha_obs, 1)
            size_obs=check_size(self.size_obs, 1)
            linestyle_obs=check_linestyle(self.linestyle_obs, 'dashed')
            marker_obs=check_marker(self.marker_obs, "d")
            label_obs=self.label_obs.get()
            if len(label_obs) == 0: label_obs="Observed"

            # General setting for plotting simulated data
            color_sim=check_color(self.color_sim, "blue")
            alpha_sim=check_alpha(self.alpha_sim, 1)
            size_sim=check_size(self.size_sim, 1)
            linestyle_sim=check_linestyle(self.linestyle_sim, 'solid')
            label_sim=self.label_sim.get()
            if len(label_sim) == 0: label_sim="Simulated"
          
            # Lable
            title=self.plot_title.get()
            xlabel=self.x_label.get()
            ylabel=self.y_label.get().strip()
            if ylabel == "": ylabel=self.globalData["target_feature_forecast_plot"]
            ylim_up=check_ylim(self.ylim_up, -9999.0)
                
            plot_window=ToplevelWindow(window_name="Plot window")
            
            # Plot
            plot_time_series(plot_window, self.globalData, key, idx, lineplot, 
                             color_obs, alpha_obs, size_obs, linestyle_obs,
                             marker_obs, label_obs, color_sim, alpha_sim, 
                             size_sim, linestyle_sim, label_sim, title, xlabel,
                             ylabel, ylim_up, forecast_period=True)
            
            
        except:
            tk.messagebox.showinfo(title="Error", 
                                   message="Cannot plot time series")
              
              




#!/usr/bin/env python

from hydroecolstm.data.read_data import read_scale_data
from hydroecolstm.model.create_model import create_model
from hydroecolstm.train.trainer import Trainer
from ray import tune, air
from pathlib import Path
from datetime import datetime
import copy
import os
import torch

# Function to train and test the model 
def run_config(config):

    # Read and split data
    data = read_scale_data(config)
    
    # Convert config to search space
    search_space = config_to_search_space(config, 
                                          data['x_train_scale'], 
                                          data['y_train_scale'],
                                          data['x_valid_scale'], 
                                          data['y_valid_scale'])
    
    if search_space["is_manual_optim"]:
        
        # Create the model
        model = create_model(config)
        
        # Train with train dataset
        trainer = Trainer(config, model)
        model = trainer.train(data['x_train_scale'], data['y_train_scale'],
                              data['x_valid_scale'], data['y_valid_scale'])
        
        # Save train loss per epoch and best train loss
        data["loss_epoch"] = trainer.loss_epoch
        data["best_train_loss"] = trainer.best_train_loss
        best_config = config
        
    else:
        
        tuner = tune.Tuner(loss,
                           param_space=search_space,
                           run_config=air.RunConfig(
                               Path(config["output_directory"][0],
                                    datetime.now().strftime("%Y%m%d_%H%M%S_%f")
                                    )
                               )
                           )
        result = tuner.fit()
        
        # Create model 
        best_result = result.get_best_result("loss", mode="min")
        best_config = best_result.config
        del best_result.config["x_train_scale"], 
        del best_result.config["y_train_scale"]
        del best_result.config["x_valid_scale"]
        del best_result.config["y_valid_scale"]
        
        model = create_model(best_result.config)
        
        with best_result.checkpoint.as_directory() as checkpoint_dir:
            model.load_state_dict(torch.load(
                os.path.join(checkpoint_dir, "model.pt"))
                )

        data["loss_epoch"] = best_result.metrics['loss_epoch']
        data["best_train_loss"] = best_result.metrics['loss']
        
    return model, data, best_config


def loss(config):
    # Create the model
    model = create_model(config)
    
    # Train with train dataset
    trainer = Trainer(config, model)
    model = trainer.train(config['x_train_scale'], config['y_train_scale'],
                          config['x_valid_scale'], config['y_valid_scale'])
        

# Convert config to tune config (search space) and add data to tune_config
def config_to_search_space(config, x_train, y_train, x_valid, y_valid):
    
    # By default this config is used for manual hyperparam optimization
    search_space = copy.deepcopy(config)
    search_space["is_manual_optim"] = True
    search_space["optim_hyperparameter"] = []

    
    # Then check if this is for automatic hyper parameter optimization
    for key in config.keys(): 
        
        # Use try because some items are not string or iterable
        try:
            if 'tune.' in config[key]:
                search_space["is_manual_optim"] = False
                search_space["optim_hyperparameter"].append(key)
                search_space[key] = eval(config[key])
        except Exception:
            pass

    # Add train and validation data to search space
    search_space["x_train_scale"] = x_train
    search_space["y_train_scale"] = y_train
    search_space["x_valid_scale"] = x_valid
    search_space["y_valid_scale"] = y_valid
    
    # Retun results of the check and also the keys of hyperparameters    
    return search_space
  
    
# Run with input with Dict[str:torch.Tensor] and torch.no_grad()
def forward_run(model, data):

    y_train_simulated_scale = model.evaluate(data['x_train_scale'])
    y_valid_simulated_scale = model.evaluate(data['x_valid_scale'])
    y_test_simulated_scale = model.evaluate(data['x_test_scale'])
    
    # Inverse scale/transform back simulated result to real scale
    data["y_train_simulated"] = data['y_scaler'].inverse(y_train_simulated_scale)
    data["y_valid_simulated"] = data['y_scaler'].inverse(y_valid_simulated_scale)
    data["y_test_simulated"] = data['y_scaler'].inverse(y_test_simulated_scale)
    
    return data



import customtkinter as ctk
import tkinter as tk
import pandas as pd
from pandastable import Table
import tkcalendar as tkc
from CTkListbox import CTkListbox
from CTkToolTip import CTkToolTip
from hydroecolstm.data.read_data import read_scale_data
   
class DataFrame(ctk.CTkScrollableFrame):
    def __init__(self, container=None, config=None, globalData=None):
        super().__init__(container)
        
        # setup the grid layout manager 
        self.config = config
        self.globalData = globalData
        self.__create_widgets()
        
    # create widgets for sidebar frame
    def __create_widgets(self): 
        
        # create tabs      
        self.tabview = ctk.CTkTabview(master=self, width = 750, border_width=1.5,
                                      fg_color = "transparent")
        self.tabview.pack(fill='both',expand=1)
        self.tabview.add("1. Load data")
        self.tabview.tab("1. Load data").grid_columnconfigure((1,1), weight=1)
        self.tabview.add("2. Filter data")
        self.tabview.tab("2. Filter data").grid_columnconfigure((1,1), weight=1)
        self.tabview.add("3. Transform data")
        self.tabview.tab("3. Transform data").grid_columnconfigure((1,1), weight=1)
        
        # ---------------------------------------------content of load data tab
        # ---------------------------------------------------------Dynamic data
        self.dynamic_label = ctk.CTkLabel(self.tabview.tab("1. Load data"), 
                                         text="1. Dynamic/time series data")
        self.dynamic_label.grid(row=0, column=0, padx = 10, pady=(5,5), sticky="w")
        
        # Select dynamic data label
        self.select_dynamic_file = ctk.CTkButton(self.tabview.tab("1. Load data"), 
                                               anchor='w', 
                                               text="Select dynamic data file", 
                                               command=self.get_dynamic_file)
        self.select_dynamic_file.grid(row=1, column=0, padx = 10, pady=(5,5), sticky="w")        
        CTkToolTip(self.select_dynamic_file, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='e',  wraplength=500, 
                   message='Click here to select dynamic (timeseries) data file. ' + 
                   'The data file must be in .csv format with header. ' +
                   'This data file must contains at least two colums, ' +
                   'one with the name <object_id> and the other column with the ' +
                   'name <time> in format yyyy-mm-dd hh:mm.The object_id column will '+ 
                   'link object in dynamic/timeseries data file with static data file. ' +
                   'Timeseries of BOTH input and target features must be in this file. ' +
                   'Template of this file can be seen in the package documentation')
        
        # display selected data file
        self.selected_dynamic_filename = ctk.CTkLabel(self.tabview.tab("1. Load data"),
                                                      text="No file was selected")
        self.selected_dynamic_filename.grid(row=2, column=0, padx = 10, pady=(5,5), sticky="w")
        
        #  display dynamic data label
        self.show_data_statistics = ctk.CTkButton(self.tabview.tab("1. Load data"), 
                                              anchor='e', 
                                              text="Display/visualize dynamic data", 
                                              command=self.display_dynamic_data)
        self.show_data_statistics.grid(row=3, column=0, padx = 10, 
                                   pady=(5,5), sticky="w")        
        
        
        # check box dynamic data
        self.check_dynamic_data = ctk.IntVar(value=0)         
        self.checkbox_dynamic_data = ctk.CTkCheckBox(self.tabview.tab("1. Load data"), 
                                               text="Check/uncheck to use/discard data",
                                               command=self.load_dynamic_data, 
                                               variable=self.check_dynamic_data)
        self.checkbox_dynamic_data.grid(row=4, column=0, padx = 10, pady=(10,10), sticky="w")
        CTkToolTip(self.checkbox_dynamic_data, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', 
                   message= 'If you check this box, data from the selected \n' +
                       'dynamic data file will be used, uncheck means data will \n' +
                       'not be used')

        # ---------------------------------------------------------Static data       
        self.static_label = ctk.CTkLabel(self.tabview.tab("1. Load data"), 
                                        text="2. Static attribute data")
        self.static_label.grid(row=0, column=2, padx = 10, pady=(5,5), sticky="w")
        
        # select static data file
        self.select_static_file = ctk.CTkButton(self.tabview.tab("1. Load data"), 
                                              anchor='e', 
                                              text="Select static data file (optional)", 
                                              command=self.get_static_file)

        self.select_static_file.grid(row=1, column=2, padx = 10, 
                                   pady=(5,5), sticky="w")
        CTkToolTip(self.select_static_file, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=500, 
                   message='Click here to select static (catchment attributes) data file. ' + 
                   'the data file must be in .csv format with header. ' + 
                   'This data file must contains a column name <object_id> ' +
                   'to connect static with dynamic data. ' +
                   'Template of this file can be seen in the package documentation')
        
        # display selected data file
        self.selected_static_filename = ctk.CTkLabel(self.tabview.tab("1. Load data"),
                                                      text="No file was selected")
        self.selected_static_filename.grid(row=2, column=2, padx = 10, pady=(5,5), sticky="w")
        
        # display visualize static data
        self.show_data = ctk.CTkButton(self.tabview.tab("1. Load data"), 
                                              anchor='e', 
                                              text="Display/visualize static data", 
                                              command=self.display_static_data)
        self.show_data.grid(row=3, column=2, padx = 10, 
                                   pady=(5,5), sticky="w")
        CTkToolTip(self.show_data, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', wraplength=500,
                   message= 'Click here to have a look at the data (using pandastable)' +
                   'here you can also visualize the data with different type of plots')
        
        # check to load/unload
        self.check_static_data = ctk.IntVar(value=0)        
        self.checkbox_static_data = ctk.CTkCheckBox(self.tabview.tab("1. Load data"), 
                                               text="Check/uncheck to use/discard data",
                                               command=self.load_static_data, 
                                               variable=self.check_static_data)
        
        self.checkbox_static_data.grid(row=4, column=2, padx = 10, pady=(5,5), sticky="w")
        CTkToolTip(self.checkbox_static_data, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', wraplength=500,
                   message= 'If you check this box, data from the selected \n' +
                       'static data file will be used, uncheck means data will \n' +
                       'not be used')
        #--------------------------------------------Content of filter data tab
        # Select input features
        self.select_input_feature_label = ctk.CTkLabel(self.tabview.tab("2. Filter data"),
                                                       text="1. Select input features")
        self.select_input_feature_label.grid(row=0, column=0, padx = 10, pady=(5,5), sticky="w")
        self.select_input_feature = CTkListbox(master=self.tabview.tab("2. Filter data"), 
                                               multiple_selection=True, border_width=1.5,
                                               text_color="black", height=175, width = 175)
        self.select_input_feature.grid(row=1, column=0, padx = 10, pady=(5,5), sticky="w")    
        CTkToolTip(self.select_input_feature_label, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', wraplength=500,
                   message= 'Select input features for the neural network. \n' +
                       'You can select as many input features as you want')
        
        # Select target features
        self.select_target_feature_label = ctk.CTkLabel(self.tabview.tab("2. Filter data"),
                                                       text="3. Select target features")
        self.select_target_feature_label.grid(row=0, column=2, padx = 10, pady=(5,5), sticky="w")
        CTkToolTip(self.select_target_feature_label, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', wraplength=500,
                   message= 'Select target features (model outputs). \n' +
                       'You can select as many target features as you want')
        self.select_target_feature = CTkListbox(master=self.tabview.tab("2. Filter data"), 
                                                multiple_selection=True, border_width=1.5,
                                               text_color="black", height=175, width = 175)
        self.select_target_feature.grid(row=1, column=2, padx = 10, pady=(5,5), sticky="w") 
        
        # Select object id
        self.object_id_label = ctk.CTkLabel(self.tabview.tab("2. Filter data"),
                                                   text="2. Select object id")
        self.object_id_label.grid(row=2, column=0, padx = 10, pady=(5,5), sticky="w")
        CTkToolTip(self.object_id_label, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', wraplength=500,
                   message= 'Select object id (basin id) that you want to put into the model. \n' +
                       'You can select as many object ids as you want')
        self.object_id = CTkListbox(master=self.tabview.tab("2. Filter data"), 
                                           multiple_selection=True, border_width=1.5,
                                           text_color="black", height=175, width = 175)
        self.object_id.grid(row=3, column=0, rowspan=7, padx = 10, pady=(5,5), sticky="w") 
        
        # Trainning period
        self.select_date_train= ctk.CTkLabel(self.tabview.tab("2. Filter data"),
                                             text="4. Training period (yyyy-mm-dd)")
        self.select_date_train.grid(row=2, column=2, padx = 10, pady=(5,5), sticky="w")    
        self.start_train = tkc.DateEntry(self.tabview.tab("2. Filter data"), 
                                         date_pattern= 'yyyy-mm-dd', width = 25,
                                         year=2000, month=1, day=1, font=ctk.CTkFont(size=14))
        self.start_train.grid(row= 3,column=2, padx=30, pady=10, sticky='e')
        self.end_train = tkc.DateEntry(self.tabview.tab("2. Filter data"), 
                                       date_pattern= 'yyyy-mm-dd', width = 25,
                                       year=2010, month=1, day=1, font=ctk.CTkFont(size=14))
        self.end_train.grid(row= 4,column=2, padx=30, pady=10, sticky='e')   
        CTkToolTip(self.select_date_train, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w', wraplength=500, 
                   message='Select starting date (upper calender box) and \n' + 
                   'ending date (lower calendar box) of the training period. \n'+
                   'Training data are used to fit model parameters.')

        # start validation calander  
        self.select_date_valid= ctk.CTkLabel(self.tabview.tab("2. Filter data"),
                                            text="5. Validation period (yyyy-mm-dd)")
        self.select_date_valid.grid(row=5, column=2, padx = 10, pady=(5,5), sticky="w")    
        self.start_valid = tkc.DateEntry(self.tabview.tab("2. Filter data"), 
                                         date_pattern= 'yyyy-mm-dd', width = 25,
                                         year=2010, month=1, day=1, font=ctk.CTkFont(size=14))
        CTkToolTip(self.select_date_valid, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w', 
                   message='Select starting date (upper calender box) and \n' + 
                   'ending date (lower calendar box) of the validtion period. \n'+
                   'Validation data ared used for early stopping.')
        self.start_valid.grid(row= 6,column=2, padx=30, pady=10, sticky='e')
        self.end_valid = tkc.DateEntry(self.tabview.tab("2. Filter data"), 
                                       date_pattern= 'yyyy-mm-dd', width = 25,
                                       year=2015, month=1, day=1, font=ctk.CTkFont(size=14))
        self.end_valid.grid(row= 7,column=2, padx=30, pady=10, sticky='e')   
        
        # start testing calander  
        self.select_date_test= ctk.CTkLabel(self.tabview.tab("2. Filter data"),
                                            text="6. Testing period (yyyy-mm-dd)")
        self.select_date_test.grid(row=8, column=2, padx = 10, pady=(5,5), sticky="w")    
        self.start_test = tkc.DateEntry(self.tabview.tab("2. Filter data"), 
                                         date_pattern= 'yyyy-mm-dd', width = 25,
                                         year=2015, month=1, day=1, font=ctk.CTkFont(size=14))
        CTkToolTip(self.select_date_test, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w', 
                   message='Select starting date (upper calender box) and \n' + 
                   'ending date (lower calendar box) of the testing period. \n' +
                   'Testing data are used for final evaluation of the model.')
        self.start_test.grid(row= 9,column=2, padx=30, pady=10, sticky='e')
        self.end_test = tkc.DateEntry(self.tabview.tab("2. Filter data"), 
                                       date_pattern= 'yyyy-mm-dd', width = 25,
                                       year=2020, month=1, day=1, font=ctk.CTkFont(size=14))
        self.end_test.grid(row= 10,column=2, padx=30, pady=10, sticky='e')   
               
        #-----------------------------------------------------3. Transform data
        self.transform_dd_label = ctk.CTkLabel(self.tabview.tab("3. Transform data"),
                                               text="1. Transformer for dynamic features")
        self.transform_dd_label.grid(row=0, column=0, pady = 3, padx = 10, sticky = "w")
        
        self.transform_dd_option = ctk.CTkOptionMenu(self.tabview.tab("3. Transform data"),
                                                   values=['MinMaxScaler', 'Z-score', 'None'],
                                                   command=self.transform_dynamic_data_option)
        self.transform_dd_option.grid(row=1, column=0, pady = 3, padx = 10, sticky = "w")
        CTkToolTip(self.transform_dd_option, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', 
                   message= 'Click here to select method for transforming dynamic data') 


        self.transform_tar_label = ctk.CTkLabel(self.tabview.tab("3. Transform data"),
                                               text="2. Transformer for target features")
        self.transform_tar_label.grid(row=4, column=0, pady = 3, padx = 10, sticky = "w")
        
        self.transform_tar_option = ctk.CTkOptionMenu(self.tabview.tab("3. Transform data"),
                                                      values=['MinMaxScaler', 'Z-score', 'None'],
                                                      command=self.transform_target_data_option) 
        self.transform_tar_option.grid(row=5, column=0, pady = (3,35), padx = 10, sticky = "w")
        CTkToolTip(self.transform_tar_option, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', 
                   message= 'Click here to select method for transforming target features') 
        
        self.transform_ss_label = ctk.CTkLabel(self.tabview.tab("3. Transform data"),
                                       text="3. Transformer for static features")
        self.transform_ss_label.grid(row=2, column=0, pady = 3, padx = 10, sticky = "w")

        self.transform_ss_option = ctk.CTkOptionMenu(self.tabview.tab("3. Transform data"),
                                                 values=['MinMaxScaler', 'Z-score', 'None'],
                                                 command=self.transform_static_data_option) 
        self.transform_ss_option.grid(row=3, column=0, pady = 3, padx = 10, sticky = "w")
        CTkToolTip(self.transform_ss_option, delay=0.1, bg_color = 'orange',
           text_color = 'black', anchor = 'w', 
           message= 'Click here to select method for transforming static data') 


        self.execute_label = ctk.CTkLabel(self.tabview.tab("3. Transform data"),
                                               text="4. Data split and transform")
        self.execute_label.grid(row=0, column=2, pady = 3, padx = 10, sticky = "w")
        
        self.execute = ctk.CTkButton(self.tabview.tab("3. Transform data"),
                                                   text="Execute",
                                                   command=self.data_split_transform) 
        self.execute.grid(row=1, column=2, pady = 3, padx = 10, sticky = "w")
        CTkToolTip(self.execute, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', 
                   message= 'Click here to split data (train, valid, test) and transform') 
        
        self.show_result_label = ctk.CTkLabel(self.tabview.tab("3. Transform data"),
                                               text="5. Display/Visualize transform data")
        self.show_result_label.grid(row=2, column=2, pady = 3, padx = 10, sticky = "w")
        
        self.show_result = ctk.CTkButton(self.tabview.tab("3. Transform data"),
                                                   text="Display/Visualize",
                                                   command=self.display_orig_trans_data) 
        self.show_result.grid(row=3, column=2, pady = 3, padx = 10, sticky = "w")
        CTkToolTip(self.show_result, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', 
                   message= 'Click here to view the original and transform data of the first object id') 

    #-----------------------------------------------------functions for widgets
    # get dynamic data file name
    def get_dynamic_file(self):
        # No file was selected when just click this button without selecting file
        self.globalData['dynamic_data_file'] = []
        self.globalData['dynamic_data'] = []
        self.globalData['dynamic_data_header'] = []
        self.globalData['object_id'] = []
        
        try:
            # get file name
            file_name = ctk.filedialog.askopenfilename(title="Select dynamic/time series data file", 
                                                       filetypes=(('csv files', '*.csv'),
                                                                  ('All files', '*.*')))
            self.globalData['dynamic_data_file'] = [file_name]
            
            # get data
            self.globalData['dynamic_data'] = pd.read_csv(file_name,
                                                          delimiter=",", header=0)
            
            # get data
            self.globalData['dynamic_data_header'] = list(self.globalData['dynamic_data'].columns)
            self.globalData['object_id'] = list(pd.unique(self.globalData['dynamic_data']["object_id"]))

            # update label
            self.selected_dynamic_filename.configure(text= '...' + file_name[-30:])
            
        except:
            None
            # update label
            self.selected_dynamic_filename.configure(text= 'No file was selected')
        
        # Show message box
        #tk.messagebox.showinfo(title="Done reading data", message="Close this box")   
      
    # show dynamic data in pandas data table
    def display_dynamic_data(self):
        try:
            self.table = Table(tk.Toplevel(self), dataframe=self.globalData['dynamic_data'], 
                               showtoolbar=True, showstatusbar=True)
            self.table.show()
        except:
            None

    # get static data file
    def get_static_file(self):
        
        self.globalData['static_data_file'] = []
        self.globalData['static_data'] = []
        self.globalData['static_data_header'] = []
        
        try:
            file_name = ctk.filedialog.askopenfilename(title="Select static data file", 
                                                  filetypes=(('csv files', '*.csv'),
                                                             ('All files', '*.*')))
            self.globalData['static_data_file'] = [file_name]
            self.globalData['static_data'] = pd.read_csv(file_name,
                                                          delimiter=",", header=0)
            self.globalData['static_data_header'] = list(self.globalData['static_data'].columns)
            
            # update label
            self.selected_static_filename.configure(text= '...' + file_name[-30:])
        except:
            None
            
            # update label
            self.selected_static_filename.configure(text= 'No file was selected')
            
    # show static data in pandas data table
    def display_static_data(self):
        try:
            self.table = Table(tk.Toplevel(self), dataframe=self.globalData['static_data'], 
                               showtoolbar=True, showstatusbar=True)
            self.table.show()  
        except:
            None    

    # Get the data
    def load_dynamic_data(self): 

        try:
            self.select_input_feature.delete(index=0, last='END')
            self.select_target_feature.delete(index=0, last='END')
            self.object_id.delete(index=0, last='END')
        except:
            None
            
        # Get the static data file name
        if self.check_dynamic_data.get() == 0:
            try: 
                del self.config['dynamic_data_file']
            except:
                None
            
        else:
            self.config['dynamic_data_file'] = self.globalData['dynamic_data_file']
            
        try:
            if (("dynamic_data_file" in self.config) & 
                (self.check_dynamic_data.get() == 1)):
                items = self.globalData['dynamic_data_header'].copy()
                object_id = self.globalData['object_id'].copy()
                 
                if (("static_data_file" in self.config) & 
                    (self.check_static_data.get() == 1)):
                    items.extend(self.globalData['static_data_header'])

                # remove columns with object_id and time
                items = [i for i in items if i != "object_id"]
                items = [i for i in items if i != "time"]
                
                for i in items:
                    self.select_input_feature.insert('END', option=i)
                    self.select_target_feature.insert('END', option=i)
                    
                for i in object_id:
                    self.object_id.insert('END', option=i)
            else:
                if (("static_data_file" in self.config) & 
                    (self.check_static_data.get() == 1)):
                    items = self.globalData['static_data_header'].copy()
                     
                    # remove columns with object_id and time
                    items = [i for i in items if i != "object_id"]
                    items = [i for i in items if i != "time"]
                    
                    for i in items:
                        self.select_input_feature.insert('END', option=i)
                        self.select_target_feature.insert('END', option=i)
        except:
            None

           
    def load_static_data(self): 
        
        # First remove content of the select input and target features box
        try:
            self.select_input_feature.delete(index=0, last='END')
            self.select_target_feature.delete(index=0, last='END')

        except:
            None
        
        # Get the static data file name
        if self.check_static_data.get() == 0:
            try: 
                del self.config['static_data_file']
            except:
                None
        else:
            self.config['static_data_file'] = self.globalData['static_data_file']
            
        # Get content for select input and target feature box
        try:
            if (("static_data_file" in self.config) & 
                (self.check_static_data.get() == 1)):
                items = self.globalData['static_data_header'].copy()

                if (("dynamic_data_file" in self.config) & 
                    (self.check_dynamic_data.get() == 1)):
                    items.extend(self.globalData['dynamic_data_header'])

                # remove columns with object_id and time
                items = [i for i in items if i != "object_id"]
                items = [i for i in items if i != "time"]
                
                for i in items:
                    self.select_input_feature.insert('END', option=i)
                    self.select_target_feature.insert('END', option=i)
            else:
                if (("dynamic_data_file" in self.config) & 
                    (self.check_dynamic_data.get() == 1)):
                    items = self.globalData['dynamic_data_header'].copy()

                    # remove columns with object_id and time
                    items = [i for i in items if i != "object_id"]
                    items = [i for i in items if i != "time"]
                    
                    for i in items:
                        self.select_input_feature.insert('END', option=i)
                        self.select_target_feature.insert('END', option=i)

        except:
            None
 
    # Get transformer 
    def transform_dynamic_data_option(self, method: str):
        self.config["scaler_input_dynamic_features"] = [method]
    
    def transform_static_data_option(self, method: str):
        self.config["scaler_input_static_features"] = [method]
        
    def transform_target_data_option(self, method: str):
        self.config["scaler_target_features"] = [method]
        
    def data_split_transform(self):
        # Train test split
        # get training and testing periods
        
        try:
            self.config['train_period'] = pd.to_datetime(
                [self.start_train.get_date(), self.end_train.get_date()],
                format = "%Y-%m-%d %H:%M")

            self.config['valid_period'] = pd.to_datetime(
                [self.start_valid.get_date(), self.end_valid.get_date()],
                format = "%Y-%m-%d %H:%M")
            
            self.config['test_period'] = pd.to_datetime(
                [self.start_test.get_date(), self.end_test.get_date()],
                format = "%Y-%m-%d %H:%M")
        except:
            pass
        
        # get input features
        all_items = self.select_input_feature.get(index='all')
        select_index = self.select_input_feature.curselection()
        input_features = [all_items[i] for i in select_index]
        
        # convert target to static and dynamic features
        if "static_data_file" in self.config:
            self.config['input_static_features'] = []
            self.config['input_dynamic_features'] = []
            
            for feature in input_features: 
                if len(self.globalData['static_data_header']) > 0:
                    if feature in self.globalData['static_data_header']:
                        self.config['input_static_features'].append(feature)
                        
                if len(self.globalData['dynamic_data_header']) > 0:
                    if feature in self.globalData['dynamic_data_header']:
                        self.config['input_dynamic_features'].append(feature)
        else:
            self.config['input_dynamic_features'] = input_features
            try: 
                del self.config['input_static_features']
            except:
                pass
        
        # get target features
        all_items = self.select_target_feature.get(index='all')
        select_index = self.select_target_feature.curselection()
        self.config['target_features'] = [all_items[i] for i in select_index]
            
        # get object id
        all_items = self.object_id.get(index='all')
        select_index = self.object_id.curselection()
        self.config['object_id'] = [all_items[i] for i in select_index]
        
        # Read data, split, and transform
        data = read_scale_data(self.config)
        self.globalData.update(data)
        
        # Show message box
        tk.messagebox.showinfo(
            title="Message box",
            message="Done test read data split (train, valid, test) and transform")
        
    def get_first_tensor(self): 

        x_train=self.globalData["x_train"].copy()
        x_train_scale=self.globalData["x_train_scale"].copy()
        

        if "input_static_features" in self.config: 
            col_names = self.config["input_dynamic_features"] +\
                self.config["input_static_features"]
        else:
            col_names = self.config["input_dynamic_features"]        

        x_train = pd.DataFrame(x_train[next(iter(x_train))])
        x_train.columns = [name + "_origignal" for name in col_names]

        x_train_scale = pd.DataFrame(x_train_scale[next(iter(x_train_scale))])
        x_train_scale.columns = [name + "_transform" for name in col_names]
     
        x_train_scale_cat = pd.concat([x_train, x_train_scale], axis=1)

        return x_train_scale_cat
        
    def display_orig_trans_data(self):
        print(self.get_first_tensor())
        try:
            self.table = Table(tk.Toplevel(self), dataframe=self.get_first_tensor(), 
                               showtoolbar=True, showstatusbar=True)
            self.table.show()
        except:
            pass        



import torch
from torch import nn

class CustomLoss(nn.Module):
    def __init__(self, loss_function:str):
        super(CustomLoss, self).__init__()

        # Dict of all available loss functions
        loss_functions = {"MSE": self.MSE, 
                          "RMSE": self.RMSE,
                          "MAE": self.MAE,
                          "NSE_complement": self.NSE_complement}
        
        # Use this loss function

        self.loss_function = loss_functions[loss_function]
    
    def forward(self, y_true:torch.Tensor, y_predict:torch.Tensor) -> torch.Tensor:
        
        # TODO: Why loss with 3D differnt with 2D Tensor
        # y_true = y_true.view(-1, y_true.size(2))
        # y_predict = y_predict.view(-1, y_predict.size(2))
                
        mask = ~torch.isnan(y_true)
        loss = self.loss_function(y_true, y_predict, mask)
            
        return loss
    
    # Mean square error
    def MSE(self, y_true:torch.Tensor, y_predict:torch.Tensor,
            mask:torch.Tensor)-> torch.Tensor: 

        # Mean square error
        loss = nn.MSELoss()
        mse = loss(y_true[mask], y_predict[mask])
        
        # Return output
        return mse
    
    # Mean absolute error
    def MAE(self, y_true:torch.Tensor, y_predict:torch.Tensor,
            mask:torch.Tensor)-> torch.Tensor: 
        
        # Mean absolute error
        loss = nn.L1Loss()
        mae = loss(y_true[mask], y_predict[mask])
        return mae      
    
    # Root mean square error
    def RMSE(self, y_true:torch.Tensor, y_predict:torch.Tensor,
            mask:torch.Tensor)-> torch.Tensor: 
        
        # Root Mean Square Error
        rmse = self.MSE(y_true, y_predict, mask)**0.5
        return rmse

    # Complement to 1 of the Nash-Sutcliffe (or 1- Nash sutcliffe)
    def NSE_complement(self, y_true:torch.Tensor, y_predict:torch.Tensor,
            mask:torch.Tensor)-> torch.Tensor: 
        
        # Sum Square Error
        sse = torch.sum((y_true[mask] - y_predict[mask])**2)
        
        # Sum of Square Difference around mean
        ssd = torch.sum((y_true[mask] - torch.mean(y_predict[mask]))**2)
        
        # Minimize loss, so output should be sse/ssd, which is 1 - NSE
        return sse/ssd


#x = CustomLoss(config["loss_function"])











# -*- coding: utf-8 -*-

import yaml
import pandas as pd

# Function to read the configurtion file
def read_config(config_file):
    
    # Open and read the configureation file
    with open(config_file, 'r') as file:
        config = yaml.safe_load(file)
    
    # All required keywords
    keys = ["object_id","input_dynamic_features", "target_features", 
           "train_period", "test_period",  "n_epochs", "learning_rate", 
           "dynamic_data_file"]
 
    # Check if required keywords are missing
    for key in keys: 
        if key not in config.keys():
            raise NameError(f"Error in configuration file, keyword '{key}' is missing")
            
    # Convert date to pandas date time
    if "train_period" in config.keys():
        config["train_period"] = pd.to_datetime(config["train_period"], 
                                                format = "%Y-%m-%d %H:%M")
    if "valid_period" in config.keys():
        config["valid_period"] = pd.to_datetime(config["valid_period"], 
                                            format = "%Y-%m-%d %H:%M")
    if "test_period" in config.keys():
        config["test_period"] = pd.to_datetime(config["test_period"], 
                                           format = "%Y-%m-%d %H:%M")
    if "forecast_period" in config.keys():
        config["forecast_period"] = pd.to_datetime(config["forecast_period"], 
                                               format = "%Y-%m-%d %H:%M")

    # Return output
    return config


from setuptools import setup
from pathlib import Path

# Read the description
readme_file = Path(__file__).absolute().parent / "README.md"
with readme_file.open("r") as fp:
    long_description = fp.read()
    
setup(
    name='hydroecolstm',
    version='0.3.4',    
    description='A python package for HydroEcological Modelling using LSTM',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/tamnva/hydroecolstm',
    author='Tam V. Nguyen',
    author_email='tamnva@gmail.com',
    packages=['hydroecolstm', 
              'hydroecolstm.data',
              'hydroecolstm.interface', 
              'hydroecolstm.utility', 
              'hydroecolstm.model',
              'hydroecolstm.train'],
    python_requires='>=3.8',
    install_requires=['pandas',
                      'numpy',
                      'torch',
                      'pandastable',
                      'tkcalendar',
                      'PyYAML',
                      'pathlib',
                      'customtkinter',
                      'CTkToolTip',
                      'CTkMessagebox',
                      'CTkListbox',
                      'matplotlib',
                      'ray[tune]'
                      ],

    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',  
        'Operating System :: POSIX :: Linux',        
        'Programming Language :: Python :: 3',
    ],

    include_package_data=True,
    package_data={'hydroecolstm': ['images/*']},
)



import torch

class EvaluationFunction():
    def __init__(self, function_name:str, nskip:int):
        
        # Dict of all available evaluation functions
        evaluation_functions = {"MSE": self.MSE, "RMSE": self.RMSE,
                                "NSE": self.NSE, "MAE": self.MAE}
        
        # Selected evaluation function
        self.eval_function = evaluation_functions[function_name]
        self.nskip = nskip
        
    def __call__(self, y_true:torch.Tensor, y_predict:torch.Tensor) -> torch.Tensor:
        
        # Get evaluation values for each basins (key), each target variables
        eval_values = {}

        for key in y_true.keys():
            eval_values[key] = self.eval_function(y_true[key][self.nskip:,],
                                                  y_predict[key][self.nskip:,])

        avg_eval_values = sum(sum(eval_values.values()))/((len(eval_values))*eval_values[next(iter(eval_values))].shape[0])
            
        return eval_values, avg_eval_values
    
    def MSE(self, ytrue:torch.Tensor, ypredict:torch.Tensor):
        mask = ~torch.isnan(ytrue)
        mse = []
        for i in range(ytrue.shape[1]):
            mse.append(torch.mean((ytrue[:,i][mask[:,i]] - ypredict[:,i][mask[:,i]])**2))
        mse = torch.stack(mse)
        return mse


    def RMSE(self, ytrue:torch.Tensor, ypredict:torch.Tensor):
        mse = self.MSE(ytrue, ypredict)
        rmse = mse**0.5
        return rmse
    
    # 1 - Nash–Sutcliffe efficiency (NSE)
    def NSE(self, ytrue:torch.Tensor, ypredict:torch.Tensor):
        mask = ~torch.isnan(ytrue)
        
        # Sum of Square Error (sse) = sum((true-predict)^2)
        # Sum of Square Difference around mean (ssd) = sum((true-mean_true)^2)
        sse = []        
        ssd = []
        for i in range(ytrue.shape[1]):
            sse.append(torch.sum((ytrue[:,i][mask[:,i]] - ypredict[:,i][mask[:,i]])**2))
            ssd.append(torch.sum((ytrue[:,i][mask[:,i]] - torch.nanmean(ytrue[:,i]))**2))
        
        # get 1 - nse, here I call it as nse
        nse = 1.0 - torch.stack(sse)/torch.stack(ssd)
        
        if torch.isnan(nse).any():
            raise ValueError("nan values found when calculating NSE - zero division")
            
        return nse
       
    def MAE(self, ytrue:torch.Tensor, ypredict:torch.Tensor):
        mask = ~torch.isnan(ytrue)
        mae = []
        for i in range(ytrue.shape[1]):
            error = ytrue[:,i][mask[:,i]] - ypredict[:,i][mask[:,i]]
            mae.append(torch.mean(torch.abs(error)))
        mae = torch.stack(mae)
        
        return mae



import numpy as np
import pandas as pd
from ray import train
import tempfile
import os
import copy
import torch
from torch.utils.data import DataLoader
from pathlib import Path
from hydroecolstm.train.custom_loss import CustomLoss
from hydroecolstm.data.custom_dataset import CustomDataset

# LSTM + Linears
class Trainer():
    def __init__(self, config, model):

        # Training parameters
        self.lr = config["learning_rate"]
        self.loss_function = CustomLoss(config["loss_function"])
        self.n_epochs = config["n_epochs"]
        self.warmup_length = config["warmup_length"]
        self.sequence_length = config["sequence_length"]
        self.batch_size = config["batch_size"]

        # Model
        self.model = model
        self.patience = config["patience"]
        self.out_dir = config["output_directory"][0]
        
        # Train and loss
        self.loss_epoch = None
        self.best_state_dict = None
        self.best_train_loss = None
        
    # Train function
    def train(self, 
              x_train: torch.Tensor, y_train: torch.Tensor,
              x_valid: torch.Tensor, y_valid: torch.Tensor,):
        
        # Optimization function
        optim = torch.optim.Adam(self.model.parameters(), lr=self.lr)
                
        # Create custom dataset
        xy_train = CustomDataset(x_train, y_train, self.sequence_length,
                                 self.warmup_length)
        xy_valid = CustomDataset(x_valid, y_valid, self.sequence_length,
                                 self.warmup_length)
        
        print("Number of iteration per epoch = ",
              int(xy_train.__len__()/self.batch_size + 0.49))
        
        # Train and valid loss per epoch
        train_loss_epoch = []
        valid_loss_epoch = []
        check_point = []
        
        # initialize early stoping
        early_stopping = EarlyStopping(patience=self.patience, verbose=False,
                                       path=self.out_dir)
        
        # Train the model
        for epoch in range(self.n_epochs):
            
            # Create batch data for each epoch
            xy_train_batch = DataLoader(xy_train, self.batch_size, shuffle=True)
            xy_valid_batch = DataLoader(xy_valid, self.batch_size, shuffle=True)
            
            # Create list to store train and valid loss per batch
            train_loss_batch = []
            valid_loss_batch = []
            
            # Set model to train mode
            self.model.train()
            
            # Loop over batches
            for x_batch, y_batch in xy_train_batch:
                
                # TODO: why with 3D tensor doesn't work, remove those 2 lines
                # -------------------------------------------------------------
                x_batch = x_batch.view(-1, x_batch.size(2))
                y_batch = y_batch.view(-1, y_batch.size(2))
                # -------------------------------------------------------------
                    
                # Get model output
                y_predict = self.model(x_batch)

                # Reset the gradients to zero
                optim.zero_grad()
                
                # Loss value    
                loss = self.loss_function(y_batch, y_predict)
                 
                # Backward prop
                loss.backward()
                    
                # Update weights and biases
                optim.step()
                
                # Save traning loss 
                train_loss_batch.append(loss.item())
                
            # Set model to eval mode (in this mode, dropout = 0, no normlization)
            self.model.eval()

            # Loop over batches
            for x_batch, y_batch in xy_valid_batch:
                
                # TODO: why with 3D tensor doesn't work, remove those 2 lines
                x_batch = x_batch.view(-1, x_batch.size(2))
                y_batch = y_batch.view(-1, y_batch.size(2))
                
                # Forward pass:
                y_predict = self.model(x_batch)
                
                # Get Loss
                loss = self.loss_function(y_batch, y_predict)
                
                # Save traning loss 
                valid_loss_batch.append(loss.item())

            # Store average loss per epoch for training and validation
            train_loss_epoch.append(np.average(train_loss_batch))
            valid_loss_epoch.append(np.average(valid_loss_batch))
            
            print(f"Epoch [{epoch+1}/{self.n_epochs}], ", 
                  f"average_train_loss = {train_loss_epoch[-1]:.8f}, ",
                  f"avearge_valid_loss = {valid_loss_epoch[-1]:.8f}")
                
            # Early stopping based on validation loss and make checkpoint
            flag = early_stopping(valid_loss_epoch[-1], self.model)
            check_point.append(flag)
            if flag: 
                self._save_check_point(train_loss_epoch, valid_loss_epoch, 
                                       check_point)
                
            if early_stopping.early_stop:
                print("Early stopping")
                break
            
        # If the model does not stops until the last epoch
        # It means that the best model is from last epoch
        if (epoch + 1) == self.n_epochs:
            print("Validation loss continue decreasing. Saving model ...")
            check_point[-1] = True
            self._save_check_point(train_loss_epoch, valid_loss_epoch, 
                                   check_point)
            
        else:
            # Load the last checkpoint with the best model
            self.model.load_state_dict(self.best_state_dict)
            
            # Set model to eval mode
            self.model.eval()
            pass
            
        self.loss_epoch = self._create_train_loss_df(train_loss_epoch, 
                                                     valid_loss_epoch, 
                                                     check_point)

        # Save loss_epoch incase of automatic hyperparam optim with tune
        #self.loss_epoch.to_csv(
        #    Path(self.out_dir, str(np.random.randint(1, 1e9)) + ".txt"),
        #    sep='\t')

        return self.model
    
    # Save intermediate result at check points
    def _save_check_point(self, train_loss_epoch, valid_loss_epoch, 
                          check_point):
        
        self.best_train_loss = train_loss_epoch[-1]
        self.best_state_dict = copy.deepcopy(self.model.state_dict())
                
        with tempfile.TemporaryDirectory() as temp_checkpoint_dir:
            
            # Save model state dict and train and valid loss
            torch.save(self.model.state_dict(),
                       os.path.join(temp_checkpoint_dir, "model.pt"))
            
            checkpoint = train.Checkpoint.from_directory(temp_checkpoint_dir)
            
            loss_epoch = self._create_train_loss_df(train_loss_epoch, 
                                                    valid_loss_epoch, 
                                                    check_point)
            
            train.report({'loss': train_loss_epoch[-1], 
                          'loss_epoch': loss_epoch},
                         checkpoint=checkpoint)
            
    # Create data frame of epoch number, train loss, valid loss
    def _create_train_loss_df(self, train_loss_epoch, valid_loss_epoch, 
                              check_point):
        
        # Initialize False list to mark best model later
        best_model = [False for i in range(len(train_loss_epoch))]
        
        # The last model is the best model
        best_model[np.where(check_point)[0][-1]] = True
        
        # Create a data frame
        loss_epoch = pd.DataFrame({
            'train_loss': train_loss_epoch,
            'valid_loss': valid_loss_epoch,
            'best_model': best_model})
        
        return loss_epoch   
# ----------------------------------------------------------------------------#
# The EarlyStopping code was copied from                                     #
# https://github.com/Bjarten/early-stopping-pytorch/blob/master/pytorchtools.py
# MIT License                                                                 #
# Copyright (c) 2018 Bjarte Mehus Sunde                                       #
#                                                                             #                                       #
# ----------------------------------------------------------------------------#
class EarlyStopping:
    """Early stops the training if validation loss doesn't improve after a given patience."""
    def __init__(self, patience=7, verbose=False, delta=0, path:str=None, trace_func=print):
        """
        Args:
            patience (int): How long to wait after last time validation loss improved.
                            Default: 7
            verbose (bool): If True, prints a message for each validation loss improvement. 
                            Default: False
            delta (float): Minimum change in the monitored quantity to qualify as an improvement.
                            Default: 0
            path (str): Path for the checkpoint to be saved to.
                            Default: 'checkpoint.pt'
            trace_func (function): trace print function.
                            Default: print            
        """
        self.patience = patience
        self.verbose = verbose
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.val_loss_min = np.inf
        self.delta = delta
        self.trace_func = trace_func
    def __call__(self, val_loss, model):

        score = -val_loss
        flag = False

        if self.best_score is None:
            self.best_score = score
            self.save_checkpoint(val_loss, model)
            flag = True
        elif score < self.best_score + self.delta:
            self.counter += 1
            if self.verbose:
                self.trace_func(f'EarlyStopping counter: {self.counter} out of {self.patience}')
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.best_score = score
            self.save_checkpoint(val_loss, model)
            self.counter = 0
            flag = True
        return flag

    def save_checkpoint(self, val_loss, model):
        '''Saves model when validation loss decrease.'''
        if self.verbose:
            self.trace_func(f'Validation loss decreased ({self.val_loss_min:.6f} --> {val_loss:.6f}).  Saving model ...')
        self.val_loss_min = val_loss



from hydroecolstm.model.lstm_linears import Lstm_Linears
from hydroecolstm.model.ea_lstm import Ea_Lstm_Linears

def create_model(config, state_dict=None):
    
    # Create the model
    if config["model_class"] == "LSTM":
        model = Lstm_Linears(config)
    else:
        model = Ea_Lstm_Linears(config)   
        
    # Assign state dict
    if state_dict is not None: 
        model.load_state_dict(state_dict)
    
    return model